import os
import sys
import toml
import unittest

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from language_pipes.cli import main 

CONFIG_PATH = "_tmp_config.toml"

class CliTests(unittest.TestCase):
    def test_empty(self):
        try:
            main([])
            self.fail("Should not succeed with no arguments")
        except:
            pass

    def test_config_file_not_exist(self):
        if os.path.exists(CONFIG_PATH):
            os.remove(CONFIG_PATH)
        try:
            main(["run", "--config", CONFIG_PATH])
            self.fail("Should not succeed with non existant config file")
        except:
            pass

    def test_config_file_empty(self):
        if os.path.exists(CONFIG_PATH):
            os.remove(CONFIG_PATH)
        with open(CONFIG_PATH, 'w', encoding='utf-8') as f:
            f.write('')
        try:
            main(["run", "--config", CONFIG_PATH])
            self.fail("Should not succeed with empty file")
        except:
            pass

    def test_config_min(self):
        if os.path.exists(CONFIG_PATH):
            os.remove(CONFIG_PATH)
        with open(CONFIG_PATH, 'w', encoding='utf-8') as f:
            toml.dump({
                "node_id": "node-1",
                "hosted_models": [{
                    "id": "meta-llama",
                    "device": "cpu",
                    "max_memory": 5
                }]
            }, f)
        main(["run", "--config", CONFIG_PATH])
    
    def test_config_file_no_id(self):
        if os.path.exists(CONFIG_PATH):
            os.remove(CONFIG_PATH)
        with open(CONFIG_PATH, 'w', encoding='utf-8') as f:
            toml.dump({
                "hosted_models": [{
                    "id": "meta-llama",
                    "device": "cpu",
                    "max_memory": 5
                }]
            }, f)
        try:
            main(["run", "--config", CONFIG_PATH])
            self.fail("Node ID should be neccessary to start program")
        except:
            pass

    def test_config_file_no_hosted(self):
        if os.path.exists(CONFIG_PATH):
            os.remove(CONFIG_PATH)
        with open(CONFIG_PATH, 'w', encoding='utf-8') as f:
            toml.dump({
                "node_id": "node-1"
            }, f)
        try:
            main(["run", "--config", CONFIG_PATH])
            self.fail("hosted_models should be necessary to start program")
        except:
            pass

    def test_min_flags(self):
        main(["run", "--node-id", "node-1", "--hosted-models", "meta-llama/Llama3.2-1B-Instruct:cpu:5"])

if __name__ == '__main__':
    unittest.main()


import os
import subprocess
import logging
from pathlib import Path
from uuid import uuid4
from typing import List, Optional, Callable
from llm_layer_collector.compute import compute_embedding, compute_head
from transformers.models.auto.tokenization_auto import AutoTokenizer
from llm_layer_collector.auto.auto_rms import AutoRMSNorm
from llm_layer_collector.auto.auto_layer import AutoDecoderLayer

import torch
from torch import tensor

from language_pipes.util.meta import MetaModel
from language_pipes.job_manager.job import ComputeStep, Job
from language_pipes.job_manager.job_data import computationStateToJobData, jobDataToComputationState
from language_pipes.llm_model.computed import ComputedData
from llm_layer_collector import LlmLayerCollector

class LlmModel:
    model_id: str
    computed: ComputedData
    process_id: str
    pipe_id: str
    collector: LlmLayerCollector

    router_id: str
    device: str
    virtual: bool

    input_embedding: Optional[torch.nn.Embedding] | bool
    layers: List[AutoDecoderLayer]
    norm: Optional[AutoRMSNorm] | bool
    head: Optional[torch.nn.Linear] | bool
    tokenizer: Callable

    start_layer: int
    end_layer: int
    loaded: bool
    num_hidden_layers: int

    def __init__(
            self,
            model_id: str,
            router_id: str,
            pipe_id: str,
            device: str,
            process_id: Optional[str] = None
    ):
        self.model_id = model_id
        self.router_id = router_id
        self.pipe_id = pipe_id
        self.loaded = False
        self.virtual = False
        self.layers = []
        self.input_embedding = None
        self.norm = None
        self.head = None
        self.start_layer = -1
        self.end_layer = -1
        self.device = device
        model_dir = os.path.join('models', self.model_id)
        if not os.path.exists(model_dir):
            self.clone_model(model_id, model_dir)
        self.collector = LlmLayerCollector(
                model_dir=os.path.join(model_dir, 'data'),
                cache_file=os.path.join(model_dir, 'cache.json'),
                device=device,
                dtype=torch.float16 
        )
        self.num_hidden_layers = self.collector.config.num_hidden_layers
        if process_id is None:
            self.process_id = str(uuid4())
        else:
            self.process_id = process_id

        self.computed = ComputedData(f'models/{model_id}')
        self.tokenizer = lambda: AutoTokenizer.from_pretrained(os.path.join(model_dir, 'data'))
        self.logger = logging.getLogger("LM NET: " + self.router_id)

    def clone_model(self, model_id: str, model_dir: str):
        repo_url = f"https://huggingface.co/{model_id}"
        clone_dir = f"{model_dir}/data"

        if not os.path.exists(clone_dir):
            Path(clone_dir).mkdir(parents=True)
        subprocess.run(["git", "clone", repo_url, clone_dir])
        subprocess.run(["git", "lfs", "install"], cwd=clone_dir, check=True)
        subprocess.run(["git", "lfs", "pull"], cwd=clone_dir, check=True)

    def load(self):
        if self.end_layer > self.num_hidden_layers:
            self.end_layer = self.num_hidden_layers - 1

        if self.input_embedding:
            self.input_embedding = self.collector.load_input_embedding(self.device)
        
        self.norm = self.collector.load_norm(self.device)
        
        if self.head:
            self.head = self.collector.load_head(self.device)
        
        if self.start_layer == -1 or self.end_layer == -1:
            self.layers = []
        else:
            self.layers = self.collector.load_layer_set(self.start_layer, self.end_layer, self.device)
        self.loaded = True
        self.virtual = False

    def print(self):
        self.logger.info(f'''
#################################
Loaded Model: {self.model_id}
Pipe ID: {self.pipe_id}
Router: {self.router_id}
Process: {self.process_id}
Embed: {self.input_embedding is not None}
Head: {self.head is not None}
Start Layer: {self.start_layer}
End Layer: {self.end_layer}
Device: {self.device}
#################################
''')

    def process_job(self, job: Job):
        self.logger.info(f'Processing job state {job.current_step.name + (", layer " + str(job.current_layer)) if job.current_step.value == 2 else ""}')
        if job.current_step == ComputeStep.TOKENIZE:
            self.tokenize(job)
        elif job.current_step == ComputeStep.EMBED:
            self.compute_embed(job)
        elif job.current_step == ComputeStep.LAYER:
            self.compute_layers(job)
        elif job.current_step == ComputeStep.NORM:
            self.compute_norm(job)
        elif job.current_step == ComputeStep.HEAD:
            self.compute_head(job)

    def tokenize(self, job: Job):
        tokenizer: AutoTokenizer = self.tokenizer()
        prompt = tokenizer.apply_chat_template([m.to_json() for m in job.messages], tokenize=False, chat_template=tokenizer.chat_template, add_generation_prompt=True)
        input_tokens = [int(t) for t in tokenizer.encode(prompt, return_tensors='pt')[0].numpy()]
        job.input_ids = input_tokens
        job.prompt_tokens = len(input_tokens)
        job.next_step()

    def raise_exception(self, msg):
        self.logger.exception(msg)
        raise Exception(msg)

    def compute_layers(
        self, 
        job: Job,
    ):
        if job.data is None:
            self.raise_exception("cannot compute layers without job data")
        comp_state = jobDataToComputationState(job.data)
        for lyr in self.layers:
            comp_state.state = lyr(comp_state)
        job.set_layer(comp_state.state, self.end_layer + 1)
        if job.current_layer == self.num_hidden_layers:
            job.next_step()
    
    def compute_embed(self, job: Job):
        if job.current_step != ComputeStep.EMBED:
            self.raise_exception('Invalid step for embedding')
        if self.input_embedding is None:
            self.raise_exception("Input Embedding must be loaded before computation")
        comp_state = compute_embedding(self.input_embedding, tensor([job.input_ids]).to(self.device), self.collector.config)
        job.data = computationStateToJobData(comp_state)
        job.next_step()

    def compute_norm(self, job: Job):
        if job.data is None or job.data.state is None:
            self.raise_exception("Cannot compute norm without job data")
        norm = self.norm(job.data.state.to(self.device))
        job.set_norm(norm)
        
    def compute_head(self, job: Job):
        if self.head is None:
            self.raise_exception("Head must be loaded before computation")
        if job.data is None or job.data.state is None:
            self.raise_exception("Cannot compute head without job data")
        head = int(compute_head(self.head, job.data.state.to(self.device))[0][0])
        job.set_output(head)
    
    def to_meta(self) -> MetaModel:
        return MetaModel(
            process_id=self.process_id,
            has_embedding=self.input_embedding is not None and self.input_embedding != False,
            has_head=self.head is not None and self.head != False,
            start_layer=self.start_layer,
            end_layer=self.end_layer,
            router_id=self.router_id,
            pipe_id=self.pipe_id,
            model_id=self.model_id,
            loaded=self.loaded,
            num_layers=self.num_hidden_layers,
            computed=ComputedData.to_meta(self.computed)
        )

    def cleanup_tensors(self):
        del self.input_embedding
        del self.norm
        del self.head
        torch.cuda.empty_cache()
        del self.layers
        torch.cuda.empty_cache()

    @staticmethod
    def from_meta(meta: MetaModel) -> 'LlmModel':
        model = LlmModel(
            model_id=meta.model_id,
            router_id=meta.router_id,
            pipe_id=meta.pipe_id,
            device='cpu',
            process_id=meta.process_id
        )
        model.loaded = meta.loaded
        model.input_embedding = None if not meta.has_embedding else True
        model.head = None if not meta.has_head else True
        model.start_layer = meta.start_layer
        model.end_layer = meta.end_layer
        model.computed = ComputedData.from_meta(meta.computed)
        model.virtual = True

        return model
    
    @staticmethod
    def from_id(model_id: str, router_id: str, pipe_id: str, device: str) -> 'LlmModel':
        model = LlmModel(model_id, router_id, pipe_id, device)
        model.computed = ComputedData(f'models/{model_id}')
        return model



import requests
import json

url = "http://127.0.0.1:6000/v1/chat/completions"

headers = {
    "Content-Type": "application/json"
}

payload = {
    "model": "Qwen/Qwen3-30B-A3B-Thinking-2507",
    "max_completion_tokens": 10,
    "messages": [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Write a haiku about distributed systems."}
    ]
}

response = requests.post(url, headers=headers, data=json.dumps(payload))

print("Status Code:", response.status_code)
print("Response JSON:", response.json())


import os
import toml
import argparse
from language_pipes.config import LpConfig
from language_pipes.util.aes import generate_aes_key

from language_pipes import LanguagePipes

VERSION = "0.4.0"

def build_parser():
    parser = argparse.ArgumentParser(description="Language Pipes CLI")
    subparsers = parser.add_subparsers(dest="command")

    # create_key command
    create_key_parser = subparsers.add_parser("create_key", help="Generate AES key")
    create_key_parser.add_argument("output_file", help="Output file for AES key")

    # run command
    run_parser = subparsers.add_parser("run", help="Run Language Pipes with config")
    run_parser.add_argument("--config", help="Path to TOML config file")
    run_parser.add_argument("--logging-level", help="Logging level (Default: INFO)")
    run_parser.add_argument("--oai-port", type=int, help="Open AI server port (Default: none)")
    run_parser.add_argument("--node-id", help="Node ID for the network (Required)")
    run_parser.add_argument("--peer-port", type=int, help="Port for peer-to-peer network (Default: 5000)")
    run_parser.add_argument("--bootstrap-address", help="Bootstrap address for network")
    run_parser.add_argument("--bootstrap-port", type=int, help="Bootstrap port for the network")
    run_parser.add_argument("--max-pipes", type=int, help="Maximum amount of pipes to host")
    run_parser.add_argument("--network-key", type=str, help="AES key to access network (Default: network.key)")
    run_parser.add_argument("--model-validation", help="Whether to validate the model weight hashes when connecting to a pipe.", default=False, action=argparse.BooleanOptionalAction)
    run_parser.add_argument("--https", help="HTTPS job communication (Default: false)", default=False, action=argparse.BooleanOptionalAction)
    run_parser.add_argument("--job-port", type=int, help="Job receiver port (Default: 5050)")
    run_parser.add_argument("--network-ip", type=str, help="IP address for the current device (only HTTPS)")
    run_parser.add_argument("--hosted-models", nargs="*", help="Hosted models in format [huggingface-id]::[device:::[max-memory] (Required)")

    return parser

def apply_overrides(data, args):
    # Environment variable mapping
    env_map = {
        "logging_level": os.getenv("LP_LOGGING_LEVEL"),
        "oai_port": os.getenv("LP_OAI_PORT"),
        "node_id": os.getenv("LP_NODE_ID"),
        "peer_port": os.getenv("LP_PEER_PORT"),
        "bootstrap_address": os.getenv("LP_BOOTSTRAP_ADDRESS"),
        "bootstrap_port": os.getenv("LP_BOOTSTRAP_PORT"),
        "network_key": os.getenv("LP_NETWORK_KEY"),
        "https": os.getenv("LP_HTTPS"),
        "model_validation": os.getenv("LP_MODEL_VALIDATION"),
        "job_port": os.getenv("LP_JOB_PORT"),
        "max_pipes": os.getenv("LP_MAX_PIPES"),
        "network_ip": os.getenv("LP_NETWORK_IP"),
        "hosted_models": os.getenv("LP_HOSTED_MODELS"),
    }

    def precedence(key, arg, d):
        if arg is not None:
            return arg
        if key in env_map and env_map[key] is not None:
            return env_map[key]
        if key in data:
            return data[key]
        return d
    
    config = {
        "logging_level": precedence("logging_level", args.logging_level, "INFO"),
        "oai_port": precedence("oai_port", args.oai_port, None),
        "node_id": precedence("node_id", args.node_id, None),
        "peer_port": int(precedence("peer_port", args.peer_port, 5000)),
        "bootstrap_address": precedence("bootstrap_address", args.bootstrap_address, None),
        "bootstrap_port": precedence("bootstrap_port", args.bootstrap_port, 5000),
        "network_key": precedence("network_key", args.network_key, "network.key"),
        "https": precedence("https", args.https, False),
        "model_validation": precedence("model_validation", args.model_validation, False),
        "job_port": int(precedence("job_port", args.job_port, 5050)),
        "max_pipes": precedence("max_pipes", args.max_pipes, 1),
        "network_ip": precedence("network_ip", args.network_ip, "127.0.0.1"),
        "hosted_models": precedence("hosted_models", args.hosted_models, None),
    }

    if config["hosted_models"] is None:
        print("Error: hosted_models param must be supplied in config")
        exit()

    if config["node_id"] is None:
        print("Error: node_id param is not supplied in config")
        exit()
    
    if config["oai_port"] is not None:
        config["oai_port"] = int(config["oai_port"])
    
    if config["bootstrap_port"] is not None:
        config["bootstrap_port"] = int(config["bootstrap_port"])
    
    config["https"] = config["https"] == "yes" or config["https"] or config["https"] == "1" or config["https"] == "true"

    hosted_models = []
    for m in config["hosted_models"]:
        if type(m) is type(''):    
            parts = m.split("::")
            if len(parts) != 3:
                raise ValueError(f"{m} is not an acceptable format for hosted_models (must be id::device::max_memory)")
            hosted_models.append({
                "id": parts[0],
                "device": parts[1],
                "max_memory": float(parts[2])
            })
        else:
            hosted_models.append(m)

    config["hosted_models"] = hosted_models

    return config

def main(argv = None):
    parser = build_parser()
    args = []
    if argv is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(argv)

    if args.command == "create_key":
        with open(args.output_file, 'wb') as f:
            f.write(generate_aes_key())
    elif args.command == "run":
        data = { }
        if args.config is not None:
            with open(args.config, "r", encoding="utf-8") as f:
                data = toml.load(f)
        data = apply_overrides(data, args)
        config = LpConfig.from_dict({
            "logging_level": data["logging_level"],
            "oai_port": data["oai_port"],
            "router": {
                "node_id": data["node_id"],
                "port": data["peer_port"],
                "https": data["https"],
                "network_ip": data["network_ip"],
                "aes_key_file": data["network_key"],
                "bootstrap_nodes": [
                    {
                        "address": data["bootstrap_address"],
                        "port": data["bootstrap_port"]
                    }
                ] if data["bootstrap_address"] is not None else []
            },
            "processor": {
                "https": data["https"],
                "max_pipes": data["max_pipes"],
                "model_validation": data["model_validation"],
                "job_port": data["job_port"],
                "hosted_models": data["hosted_models"]
            }
        })

        return LanguagePipes(VERSION, config)
    else:
        parser.print_usage()
        exit(1)

if __name__ == "__main__":
    main()
